"""Simulator initialization."""

from whynot.simulators import civil_violence
from whynot.simulators import dice
from whynot.simulators import hiv
from whynot.simulators import lalonde
from whynot.simulators import lotka_volterra
from whynot.simulators import opioid
from whynot.simulators import schelling
from whynot.simulators import simpact_hiv
from whynot.simulators import world2
from whynot.simulators import world3
