"""Simpact HIV initialization."""

from whynot.simulators.simpact_hiv import configs
from whynot.simulators.simpact_hiv.simulator import simulate
from whynot.simulators.simpact_hiv.experiments import *
